Package: data.table
Version: 1.18.6.1
Title: Extension of `data.frame`
Depends: R (>= 3.4.0)
Imports: methods
Suggests: bit64 (>= 4.0.0), bit (>= 4.0.4), R.utils (>= 2.13.0), xts,
        zoo (>= 1.8-1), yaml, knitr, markdown
Description: Fast aggregation of large data (e.g. 100GB in RAM), fast ordered joins, fast add/modify/delete of columns by group using no copies at all, list columns, friendly and fast character-separated-value read/write. Offers a natural and flexible syntax, for faster development.
License: MPL-2.0 | file LICENSE
URL: https://r-datatable.com, https://Rdatatable.gitlab.io/data.table,
        https://github.com/Rdatatable/data.table
BugReports: https://github.com/Rdatatable/data.table/issues
VignetteBuilder: knitr
Encoding: UTF-8
ByteCompile: TRUE
Authors@R: c(
  person("Tyson","Barrett",        role=c("aut","cre"), email="t.barrett88@gmail.com", comment = c(ORCID="0000-0002-2137-1391")),
  person("Matt","Dowle",           role="aut",          email="mattjdowle@gmail.com"),
  person("Arun","Srinivasan",      role="aut",          email="asrini@pm.me"),
  person("Jan","Gorecki",          role="aut",          email="j.gorecki@wit.edu.pl"),
  person("Michael","Chirico",      role="aut",          email="michaelchirico4@gmail.com", comment = c(ORCID="0000-0003-0787-087X")),
  person("Toby","Hocking",         role="aut",          email="toby.hocking@r-project.org", comment = c(ORCID="0000-0002-3146-0865")),
  person("Benjamin","Schwendinger",role="aut", comment = c(ORCID="0000-0003-3315-8114")),
  person("Ivan", "Krylov",         role="aut",          email="ikrylov@disroot.org",   comment = c(ORCID="0000-0002-0172-3812")),
  person("Pasha","Stetsenko",      role="ctb"),
  person("Tom","Short",            role="ctb"),
  person("Steve","Lianoglou",      role="ctb"),
  person("Eduard","Antonyan",      role="ctb"),
  person("Markus","Bonsch",        role="ctb"),
  person("Hugh","Parsonage",       role="ctb"),
  person("Scott","Ritchie",        role="ctb"),
  person("Kun","Ren",              role="ctb"),
  person("Xianying","Tan",         role="ctb"),
  person("Rick","Saporta",         role="ctb"),
  person("Otto","Seiskari",        role="ctb"),
  person("Xianghui","Dong",        role="ctb"),
  person("Michel","Lang",          role="ctb"),
  person("Watal","Iwasaki",        role="ctb"),
  person("Seth","Wenchel",         role="ctb"),
  person("Karl","Broman",          role="ctb"),
  person("Tobias","Schmidt",       role="ctb"),
  person("David","Arenburg",       role="ctb"),
  person("Ethan","Smith",          role="ctb"),
  person("Francois","Cocquemas",   role="ctb"),
  person("Matthieu","Gomez",       role="ctb"),
  person("Philippe","Chataignon",  role="ctb"),
  person("Nello","Blaser",         role="ctb"),
  person("Dmitry","Selivanov",     role="ctb"),
  person("Andrey","Riabushenko",   role="ctb"),
  person("Cheng","Lee",            role="ctb"),
  person("Declan","Groves",        role="ctb"),
  person("Daniel","Possenriede",   role="ctb"),
  person("Felipe","Parages",       role="ctb"),
  person("Denes","Toth",           role="ctb"),
  person("Mus","Yaramaz-David",    role="ctb"),
  person("Ayappan","Perumal",      role="ctb"),
  person("James","Sams",           role="ctb"),
  person("Martin","Morgan",        role="ctb"),
  person("Michael","Quinn",        role="ctb"),
  person(given="@javrucebo",       role="ctb", comment="GitHub user"),
  person("Marc","Halperin",        role="ctb"),
  person("Roy","Storey",           role="ctb"),
  person("Manish","Saraswat",      role="ctb"),
  person("Morgan","Jacob",         role="ctb"),
  person("Michael","Schubmehl",    role="ctb"),
  person("Davis","Vaughan",        role="ctb"),
  person("Leonardo","Silvestri",   role="ctb"),
  person("Jim","Hester",           role="ctb"),
  person("Anthony","Damico",       role="ctb"),
  person("Sebastian","Freundt",    role="ctb"),
  person("David","Simons",         role="ctb"),
  person("Elliott","Sales de Andrade", role="ctb"),
  person("Cole","Miller",          role="ctb"),
  person("Jens Peder","Meldgaard", role="ctb"),
  person("Vaclav","Tlapak",        role="ctb"),
  person("Kevin","Ushey",          role="ctb"),
  person("Dirk","Eddelbuettel",    role="ctb"),
  person("Tony","Fischetti",       role="ctb"),
  person("Ofek","Shilon",          role="ctb"),
  person("Vadim","Khotilovich",    role="ctb"),
  person("Hadley","Wickham",       role="ctb"),
  person("Bennet","Becker",        role="ctb"),
  person("Kyle","Haynes",          role="ctb"),
  person("Boniface Christian","Kamgang", role="ctb"),
  person("Olivier","Delmarcell",   role="ctb"),
  person("Josh","O'Brien",         role="ctb"),
  person("Dereck","Mezquita",      role="ctb", comment = c(ORCID="0000-0002-9307-6762")),
  person("Michael","Czekanski",    role="ctb"),
  person("Dmitry", "Shemetov",     role="ctb"),
  person("Nitish", "Jha",          role="ctb"),
  person("Joshua", "Wu",           role="ctb"),
  person("Iago", "Giné-Vázquez",   role="ctb"),
  person("Anirban", "Chetia",      role="ctb"),
  person("Doris", "Amoakohene",    role="ctb"),
  person("Angel", "Feliz",         role="ctb"),
  person("Michael","Young",        role="ctb"),
  person("Mark", "Seeto",          role="ctb"),
  person("Philippe", "Grosjean",   role="ctb"),
  person("Vincent", "Runge",       role="ctb"),
  person("Christian", "Wia",       role="ctb"),
  person("Elise", "Maigné",        role="ctb"),
  person("Vincent", "Rocher",      role="ctb"),
  person("Vijay", "Lulla",         role="ctb"),
  person("Aljaž", "Sluga",         role="ctb"),
  person("Bill", "Evans",          role="ctb"),
  person("Reino", "Bruner",        role="ctb"),
  person(given="@badasahog",       role="ctb", comment="GitHub user"),
  person("Vinit", "Thakur",        role="ctb"),
  person("Mukul", "Kumar",         role="ctb"),
  person("Ildikó", "Czeller",      role="ctb"),
  person("Manmita", "Das",         role="ctb"),
  person("Tarun", "Thammisetty",   role="ctb"),
  person("Marco", "Colombo",       role="ctb", comment = c(ORCID = "0000-0001-6672-0623")),
  person("Tim", "Taylor",          role="ctb", comment = c(ORCID = "0000-0002-8587-7113"))
  )
NeedsCompilation: yes
Packaged: 2026-08-22 18:37:47 UTC; a00932230
Author: Tyson Barrett [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-2137-1391>),
  Matt Dowle [aut],
  Arun Srinivasan [aut],
  Jan Gorecki [aut],
  Michael Chirico [aut] (ORCID: <https://orcid.org/0000-0003-0787-087X>),
  Toby Hocking [aut] (ORCID: <https://orcid.org/0000-0002-3146-0865>),
  Benjamin Schwendinger [aut] (ORCID:
    <https://orcid.org/0000-0003-3315-8114>),
  Ivan Krylov [aut] (ORCID: <https://orcid.org/0000-0002-0172-3812>),
  Pasha Stetsenko [ctb],
  Tom Short [ctb],
  Steve Lianoglou [ctb],
  Eduard Antonyan [ctb],
  Markus Bonsch [ctb],
  Hugh Parsonage [ctb],
  Scott Ritchie [ctb],
  Kun Ren [ctb],
  Xianying Tan [ctb],
  Rick Saporta [ctb],
  Otto Seiskari [ctb],
  Xianghui Dong [ctb],
  Michel Lang [ctb],
  Watal Iwasaki [ctb],
  Seth Wenchel [ctb],
  Karl Broman [ctb],
  Tobias Schmidt [ctb],
  David Arenburg [ctb],
  Ethan Smith [ctb],
  Francois Cocquemas [ctb],
  Matthieu Gomez [ctb],
  Philippe Chataignon [ctb],
  Nello Blaser [ctb],
  Dmitry Selivanov [ctb],
  Andrey Riabushenko [ctb],
  Cheng Lee [ctb],
  Declan Groves [ctb],
  Daniel Possenriede [ctb],
  Felipe Parages [ctb],
  Denes Toth [ctb],
  Mus Yaramaz-David [ctb],
  Ayappan Perumal [ctb],
  James Sams [ctb],
  Martin Morgan [ctb],
  Michael Quinn [ctb],
  @javrucebo [ctb] (GitHub user),
  Marc Halperin [ctb],
  Roy Storey [ctb],
  Manish Saraswat [ctb],
  Morgan Jacob [ctb],
  Michael Schubmehl [ctb],
  Davis Vaughan [ctb],
  Leonardo Silvestri [ctb],
  Jim Hester [ctb],
  Anthony Damico [ctb],
  Sebastian Freundt [ctb],
  David Simons [ctb],
  Elliott Sales de Andrade [ctb],
  Cole Miller [ctb],
  Jens Peder Meldgaard [ctb],
  Vaclav Tlapak [ctb],
  Kevin Ushey [ctb],
  Dirk Eddelbuettel [ctb],
  Tony Fischetti [ctb],
  Ofek Shilon [ctb],
  Vadim Khotilovich [ctb],
  Hadley Wickham [ctb],
  Bennet Becker [ctb],
  Kyle Haynes [ctb],
  Boniface Christian Kamgang [ctb],
  Olivier Delmarcell [ctb],
  Josh O'Brien [ctb],
  Dereck Mezquita [ctb] (ORCID: <https://orcid.org/0000-0002-9307-6762>),
  Michael Czekanski [ctb],
  Dmitry Shemetov [ctb],
  Nitish Jha [ctb],
  Joshua Wu [ctb],
  Iago Giné-Vázquez [ctb],
  Anirban Chetia [ctb],
  Doris Amoakohene [ctb],
  Angel Feliz [ctb],
  Michael Young [ctb],
  Mark Seeto [ctb],
  Philippe Grosjean [ctb],
  Vincent Runge [ctb],
  Christian Wia [ctb],
  Elise Maigné [ctb],
  Vincent Rocher [ctb],
  Vijay Lulla [ctb],
  Aljaž Sluga [ctb],
  Bill Evans [ctb],
  Reino Bruner [ctb],
  @badasahog [ctb] (GitHub user),
  Vinit Thakur [ctb],
  Mukul Kumar [ctb],
  Ildikó Czeller [ctb],
  Manmita Das [ctb],
  Tarun Thammisetty [ctb],
  Marco Colombo [ctb] (ORCID: <https://orcid.org/0000-0001-6672-0623>),
  Tim Taylor [ctb] (ORCID: <https://orcid.org/0000-0002-8587-7113>)
Maintainer: Tyson Barrett <t.barrett88@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-24 08:40:13 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-08-24 12:53:03 UTC; unix
Archs: data_table.so.dSYM
