```{r test_id, message=FALSE, results="show", echo=TRUE, warning=FALSE}
require(data.table)              # print?
DT = data.table(x=1:3, y=4:6)    # no
DT                               # yes
DT[, z := 7:9]                   # no
print(DT[, z := 10:12])          # yes
if (1 < 2) DT[, a := 1L]         # no
DT                               # yes
```
Some text.

